# Generated from D:/Workspaces/pycharm-workspace/Qrogue/qrogue/management/save_grammar/SaveData.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .SaveDataParser import SaveDataParser
else:
    from SaveDataParser import SaveDataParser

# This class defines a complete listener for a parse tree produced by SaveDataParser.
class SaveDataListener(ParseTreeListener):

    # Enter a parse tree produced by SaveDataParser#start.
    def enterStart(self, ctx:SaveDataParser.StartContext):
        pass

    # Exit a parse tree produced by SaveDataParser#start.
    def exitStart(self, ctx:SaveDataParser.StartContext):
        pass


    # Enter a parse tree produced by SaveDataParser#date_time.
    def enterDate_time(self, ctx:SaveDataParser.Date_timeContext):
        pass

    # Exit a parse tree produced by SaveDataParser#date_time.
    def exitDate_time(self, ctx:SaveDataParser.Date_timeContext):
        pass


    # Enter a parse tree produced by SaveDataParser#duration.
    def enterDuration(self, ctx:SaveDataParser.DurationContext):
        pass

    # Exit a parse tree produced by SaveDataParser#duration.
    def exitDuration(self, ctx:SaveDataParser.DurationContext):
        pass


    # Enter a parse tree produced by SaveDataParser#inventory.
    def enterInventory(self, ctx:SaveDataParser.InventoryContext):
        pass

    # Exit a parse tree produced by SaveDataParser#inventory.
    def exitInventory(self, ctx:SaveDataParser.InventoryContext):
        pass


    # Enter a parse tree produced by SaveDataParser#gates.
    def enterGates(self, ctx:SaveDataParser.GatesContext):
        pass

    # Exit a parse tree produced by SaveDataParser#gates.
    def exitGates(self, ctx:SaveDataParser.GatesContext):
        pass


    # Enter a parse tree produced by SaveDataParser#gate.
    def enterGate(self, ctx:SaveDataParser.GateContext):
        pass

    # Exit a parse tree produced by SaveDataParser#gate.
    def exitGate(self, ctx:SaveDataParser.GateContext):
        pass


    # Enter a parse tree produced by SaveDataParser#combined_gate.
    def enterCombined_gate(self, ctx:SaveDataParser.Combined_gateContext):
        pass

    # Exit a parse tree produced by SaveDataParser#combined_gate.
    def exitCombined_gate(self, ctx:SaveDataParser.Combined_gateContext):
        pass


    # Enter a parse tree produced by SaveDataParser#background_gate.
    def enterBackground_gate(self, ctx:SaveDataParser.Background_gateContext):
        pass

    # Exit a parse tree produced by SaveDataParser#background_gate.
    def exitBackground_gate(self, ctx:SaveDataParser.Background_gateContext):
        pass


    # Enter a parse tree produced by SaveDataParser#levels.
    def enterLevels(self, ctx:SaveDataParser.LevelsContext):
        pass

    # Exit a parse tree produced by SaveDataParser#levels.
    def exitLevels(self, ctx:SaveDataParser.LevelsContext):
        pass


    # Enter a parse tree produced by SaveDataParser#level.
    def enterLevel(self, ctx:SaveDataParser.LevelContext):
        pass

    # Exit a parse tree produced by SaveDataParser#level.
    def exitLevel(self, ctx:SaveDataParser.LevelContext):
        pass


    # Enter a parse tree produced by SaveDataParser#score.
    def enterScore(self, ctx:SaveDataParser.ScoreContext):
        pass

    # Exit a parse tree produced by SaveDataParser#score.
    def exitScore(self, ctx:SaveDataParser.ScoreContext):
        pass


    # Enter a parse tree produced by SaveDataParser#value.
    def enterValue(self, ctx:SaveDataParser.ValueContext):
        pass

    # Exit a parse tree produced by SaveDataParser#value.
    def exitValue(self, ctx:SaveDataParser.ValueContext):
        pass


    # Enter a parse tree produced by SaveDataParser#unlocks.
    def enterUnlocks(self, ctx:SaveDataParser.UnlocksContext):
        pass

    # Exit a parse tree produced by SaveDataParser#unlocks.
    def exitUnlocks(self, ctx:SaveDataParser.UnlocksContext):
        pass


    # Enter a parse tree produced by SaveDataParser#unlock.
    def enterUnlock(self, ctx:SaveDataParser.UnlockContext):
        pass

    # Exit a parse tree produced by SaveDataParser#unlock.
    def exitUnlock(self, ctx:SaveDataParser.UnlockContext):
        pass


    # Enter a parse tree produced by SaveDataParser#achievements.
    def enterAchievements(self, ctx:SaveDataParser.AchievementsContext):
        pass

    # Exit a parse tree produced by SaveDataParser#achievements.
    def exitAchievements(self, ctx:SaveDataParser.AchievementsContext):
        pass


    # Enter a parse tree produced by SaveDataParser#achievement.
    def enterAchievement(self, ctx:SaveDataParser.AchievementContext):
        pass

    # Exit a parse tree produced by SaveDataParser#achievement.
    def exitAchievement(self, ctx:SaveDataParser.AchievementContext):
        pass


    # Enter a parse tree produced by SaveDataParser#integer.
    def enterInteger(self, ctx:SaveDataParser.IntegerContext):
        pass

    # Exit a parse tree produced by SaveDataParser#integer.
    def exitInteger(self, ctx:SaveDataParser.IntegerContext):
        pass


    # Enter a parse tree produced by SaveDataParser#complex_number.
    def enterComplex_number(self, ctx:SaveDataParser.Complex_numberContext):
        pass

    # Exit a parse tree produced by SaveDataParser#complex_number.
    def exitComplex_number(self, ctx:SaveDataParser.Complex_numberContext):
        pass



del SaveDataParser